import logging
import os
import re
from pathlib import Path

import pytest
from pyscaffold import __version__ as pyscaffold_version
from pyscaffold.api import NO_CONFIG, create_project
from pyscaffold.cli import parse_args, run
from pyscaffold.templates import get_template

from pyscaffoldext.django.extension import Django, DjangoAdminNotInstalled

PROJ_NAME = "proj"
DJANGO_FILES = ["proj/manage.py", "proj/src/proj/wsgi.py", "proj/src/proj/__main__.py"]

FLAG = Django().flag


@pytest.mark.slow
def test_create_project_with_django(tmpfolder):
    # Given options with the django extension,
    opts = dict(
        project_path=PROJ_NAME,
        name=PROJ_NAME,
        package=PROJ_NAME,
        extensions=[Django()],
        version=pyscaffold_version,
        config_files=NO_CONFIG,
    )
    # NO_CONFIG: avoid extra config from dev's machine interference

    # when the project is created,
    create_project(opts)

    # then django files should exist
    for path in DJANGO_FILES:
        assert Path(path).exists()
    # and also overwritable pyscaffold files (with the exact contents)
    existing = (tmpfolder / PROJ_NAME / "setup.py").read_text()
    assert existing == get_template("setup_py").safe_substitute(opts)


def test_pretend_create_project_with_django(tmpfolder, isolated_log):
    # Given options with the django extension,
    isolated_log.set_level(logging.INFO)
    opts = parse_args([PROJ_NAME, "--no-config", "--pretend", FLAG])
    # --no-config: avoid extra config from dev's machine interference

    # when the project is created,
    create_project(opts)

    # then files should exist
    assert not Path(PROJ_NAME).exists()
    for path in DJANGO_FILES:
        assert not Path(path).exists()

    # but activities should be logged
    logs = isolated_log.text
    assert re.search(r"run.+django", logs)


def test_create_project_without_django(tmpfolder):
    # Given options without the django extension,
    opts = dict(project_path=PROJ_NAME, config_files=NO_CONFIG)
    # NO_CONFIG: avoid extra config from dev's machine interference

    # when the project is created,
    create_project(opts)

    # then django files should not exist
    for path in DJANGO_FILES:
        assert not Path(path).exists()


def test_create_project_no_django(tmpfolder, nodjango_admin_mock):
    # Given options with the django extension,
    # but without django-admin being installed,
    opts = dict(project_path=PROJ_NAME, extensions=[Django()], config_files=NO_CONFIG)
    # NO_CONFIG: avoid extra config from dev's machine interference

    # when the project is created,
    # then an exception should be raised.
    with pytest.raises(DjangoAdminNotInstalled):
        create_project(opts)


@pytest.mark.slow
def test_cli_with_django(tmpfolder):
    # Given the command line with the django option,
    args = ["--no-config", FLAG, PROJ_NAME]
    # --no-config: avoid extra config from dev's machine interference

    # when pyscaffold runs,
    run(args)

    # then django files should exist
    for path in DJANGO_FILES:
        assert Path(path).exists()


def test_cli_without_django(tmpfolder):
    # Given the command line without the django option,
    args = ["-vv", "--no-config", PROJ_NAME]
    # --no-config: avoid extra config from dev's machine interference

    # when pyscaffold runs,
    run(args)

    # then django files should not exist
    for path in DJANGO_FILES:
        assert not Path(path).exists()


def test_cli_with_django_and_update(tmpfolder, capfd):
    # Given a project exists
    create_project(project_path=PROJ_NAME, config_files=NO_CONFIG)
    # NO_CONFIG: avoid extra config from dev's machine interference

    # when the project is updated
    # with the django extension,
    run([PROJ_NAME, "--no-config", "--update", FLAG])
    # --no-config: avoid extra config from dev's machine interference

    # then a warning should be displayed
    out, err = capfd.readouterr()
    out_err = out + err
    try:
        assert "external tools" in out_err
        assert "not supported" in out_err
        assert "will be ignored" in out_err
    except AssertionError:
        if os.name == "nt":
            pytest.skip("pytest is having problems to capture stderr/stdout on Windows")
        else:
            raise



"""Place for fixtures and configuration that will be used in most of the tests.
A nice option is to put your ``autouse`` fixtures here.
Functions that can be imported and re-used are more suitable for the ``helpers`` file.
"""
import logging
import os
from pathlib import Path
from tempfile import mkdtemp

import pytest
from pyscaffold.exceptions import ShellCommandException
from pyscaffold.log import ReportFormatter

from .helpers import rmpath, uniqstr


@pytest.fixture
def tmpfolder(tmp_path):
    old_path = os.getcwd()
    new_path = mkdtemp(dir=str(tmp_path))
    os.chdir(new_path)
    try:
        yield Path(new_path)
    finally:
        os.chdir(old_path)
        rmpath(new_path)


@pytest.fixture
def nodjango_admin_mock(monkeypatch):
    def raise_error(*_):
        raise ShellCommandException("No django_admin mock!")

    monkeypatch.setattr("pyscaffoldext.django.extension.django_admin", raise_error)
    yield


@pytest.fixture
def isolated_log(request, monkeypatch, caplog):
    """See isolated_logger in pyscaffold/tests/conftest.py to see why this fixture
    is important to guarantee tests checking logs work as expected.
    This just work for multiprocess environments, not multithread.
    """
    if "original_logger" in request.keywords:
        yield caplog
        return

    # Get a fresh new logger, not used anywhere
    raw_logger = logging.getLogger(uniqstr())
    raw_logger.setLevel(logging.NOTSET)
    new_handler = logging.StreamHandler()

    patches = {
        "propagate": True,  # <- needed for caplog
        "nesting": 0,
        "wrapped": raw_logger,
        "handler": new_handler,
        "formatter": ReportFormatter(),
    }
    for key, value in patches.items():
        monkeypatch.setattr(f"pyscaffold.log.logger.{key}", value)

    try:
        yield caplog
    finally:
        new_handler.close()



import sys
from contextlib import suppress
from glob import glob
from pathlib import Path
from subprocess import CalledProcessError

import pytest
from pyscaffold import file_system as fs
from pyscaffold import shell
from pyscaffold.identification import underscore

from pyscaffoldext.django.extension import Django

from .helpers import PYTHON, merge_env, run

FLAG = Django().flag
PUTUP = shell.get_executable("putup")
PIP = shell.get_executable("pip")


def is_venv():
    """Check if the tests are running inside a venv"""
    return hasattr(sys, "real_prefix") or (
        hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
    )


@pytest.fixture(autouse=True)
def cwd(tmpfolder):
    """Guarantee a blank folder as workspace"""
    yield tmpfolder


def chdir(path):
    return fs.chdir(fs.create_directory(path))


def test_ensure_inside_test_venv():
    # This is a METATEST
    # Here we ensure `putup` is installed inside tox or
    # a local virtualenv (pytest-runner), so we know we are testing the correct
    # version of pyscaffold and not one the devs installed to use in other
    # projects
    assert PUTUP
    assert PIP
    assert ".tox" in PUTUP or is_venv()
    assert Path(PUTUP).parent == Path(PIP).parent


def test_django_generates_files(tmpfolder):
    # Given pyscaffold is installed,
    # when we call putup with extensions
    name = "myproj"
    run(PUTUP, "--no-config", FLAG, name)
    # --no-config: avoid extra config from dev's machine interference
    with chdir(tmpfolder / name):
        # then special files should be created
        assert Path("manage.py").exists()
        assert Path("src/myproj/__main__.py").exists()
        assert not Path("myproj").exists()


TASKS = [
    "test",
    "clearsessions",
    "showmigrations",
    "createsuperuser --username admin --email admin@localhost --no-input",
]


def remove_eventual_package(name):
    with suppress(Exception, CalledProcessError):
        run(f"{PIP} uninstall --yes {name}")


RND_NAME1 = "pkg82b3f28b-2811"
RND_NAME2 = "pkg9402c0de-c719"
# ^  Fixed random values.
#    If we use random values for parametrization pytest-xdist gets confused
#    and returns an error.


@pytest.mark.slow
@pytest.mark.system
@pytest.mark.parametrize(
    "name,command",
    [
        (RND_NAME1, f"{PYTHON} manage.py"),
        (lambda x: (x, f"{PYTHON} -m " + underscore(x)))(RND_NAME2),
    ],
)
def test_manage_py_runs_nicely(tmpfolder, name, command):
    # Given we have a project generated with putup --django pkg
    try:
        remove_eventual_package(name)
        run(PUTUP, "--no-config", FLAG, name)
        with chdir(tmpfolder / name):
            print(list(glob("*")))
            print(list(Path("src").glob("*/*")))
            print(Path("manage.py").read_text())
            env = merge_env(PYTHONPATH=str(Path("src").resolve()))
            # ^  required for `python -m` to work, without installing the package

            # when we call manage.py or python -m djangotest
            run(f"{command} migrate", env=env)
            # then it should create a sqlite3 file in the right place
            assert Path("db.sqlite3").exists()
            assert not Path("src/db.sqlite3").exists()
            # and everything should be fine
            for task in TASKS:
                run(f"{command} {task}", env=env)
    finally:
        remove_eventual_package(name)



"""
Extension that creates a base structure for the project using django-admin.
from typing import List
"""

# This file was transfered from the main PyScaffold repository using
# ``git filter-branch``, and therefore might have lost parts of its
# commit history.
# Please refer to ``pyscaffold`` if that is needed.

import re
import stat
from functools import partial
from pathlib import Path
from typing import List

from pyscaffold import file_system as fs
from pyscaffold.actions import Action, ActionParams, ScaffoldOpts, Structure
from pyscaffold.extensions import Extension
from pyscaffold.log import logger
from pyscaffold.operations import add_permissions
from pyscaffold.shell import get_command
from pyscaffold.structure import merge, reify_content, resolve_leaf
from pyscaffold.templates import get_template

from . import templates

django_admin = get_command("django-admin")
template = partial(get_template, relative_to=templates)

UPDATE_WARNING = (
    "Updating code generated using external tools is not "
    "supported. The extension `django` will be ignored, only "
    "changes in PyScaffold core features will take place."
)


class Django(Extension):
    """Generate Django project files"""

    persist = False

    def activate(self, actions: List[Action]) -> List[Action]:
        """Activate extension. See :obj:`pyscaffold.extension.Extension.activate`."""
        actions = self.register(actions, enforce_options, after="get_default_options")
        actions = self.register(actions, create_django)
        return self.register(actions, instruct_user, before="report_done")


def enforce_options(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Make sure options reflect the Django usage.
    See :obj:`pyscaffold.actions.Action`.
    """
    opts["force"] = True
    opts.setdefault("requirements", []).append("django")

    return struct, opts


def create_django(struct: Structure, opts: ScaffoldOpts) -> ActionParams:
    """Creates a standard Django project with django-admin.
    See :obj:`pyscaffold.actions.Action`.
    Raises:
        :obj:`RuntimeError`: raised if django-admin is not installed
    """
    if opts.get("update"):
        logger.warning(UPDATE_WARNING)
        return struct, opts

    try:
        django_admin("--version")
    except Exception as e:
        raise DjangoAdminNotInstalled from e

    pretend = opts.get("pretend")
    project_path = Path(opts["project_path"])
    pkg_name = opts["package"]
    fs.create_directory(project_path, pretend=pretend)
    django_admin("startproject", pkg_name, str(project_path), log=True, pretend=pretend)

    src_dir = project_path / "src"
    pkg_dir = src_dir / pkg_name
    orig_dir = project_path / pkg_name

    if not pretend:
        src_dir.mkdir(exist_ok=True)
        orig_dir.rename(pkg_dir)
    logger.report("move", orig_dir, target=pkg_dir)

    manage = project_path / "manage.py"
    main = pkg_dir / "__main__.py"

    if not pretend:
        manage.rename(main)
    logger.report("move", manage, target=main)

    settings = pkg_dir / "settings.py"
    replace_default_database(logger, settings, pretend=pretend)

    contents, file_op = resolve_leaf(struct[".gitignore"])
    gitignore = reify_content(contents, opts) + "{}\n\n# Django\n/*.sqlite3\n"

    files: Structure = {
        ".gitignore": (gitignore, file_op),
        "manage.py": (template("manage"), add_permissions(stat.S_IXUSR)),
    }

    return merge(struct, files), opts


def instruct_user(struct, opts):
    logger.warning(
        "\nDjango is used to create web applications while PyScaffold makes it "
        "easy to create re-usable Python (pip) packages, such as libraries.\n"
        "There is nothing wrong with trying to distribute your web application "
        "as an installable package, but you have to be aware about the changes "
        "in mindset. Please check the official docs in\n\n"
        "\thttps://pyscaffold.org/projects/django\n\n"
        "for more information.\n"
    )

    return struct, opts


PATTERN = re.compile(r"BASE_DIR\s*/\s*['\"]db\.sqlite3['\"]")
REPLACEMENT = 'BASE_DIR.parent / "db.sqlite3"'


def replace_default_database(
    logger, file_path, pattern=PATTERN, replacement=REPLACEMENT, pretend=False
):
    exception = DjangoVersionMightBeUnsupported(
        "Failed attempt to replace the default sqlite3 database file with "
        f"{REPLACEMENT} in {file_path}."
    )

    try:
        if not pretend:
            text = file_path.read_text()
            replaced, n = pattern.subn(replacement, text)
            if n < 1:
                # No substitution was made, there is something wrong with the
                # assumptions on how the file should be.
                raise SystemError(text)
            file_path.write_text(replaced)
        logger.report("replace", f"default database in {file_path}")
    except PyScaffoldDjangoError:
        raise
    except Exception as ex:
        raise exception from ex


class PyScaffoldDjangoError(RuntimeError):
    """Base class for all the exceptions in this package"""


class DjangoVersionMightBeUnsupported(PyScaffoldDjangoError):
    """The files generated by django-admin are different from expected."""

    EXTRA = (
        "\nA possible reason for that is that a new version of Django is "
        "incompatible with pyscaffoldext-django.\n"
        "Please visit the documentation in https://pyscaffold.org/projects/django "
        "and follow the steps to integrate django manually.\n"
        "Please also consider submitting a pull request with the fix."
    )

    def __init__(self, message=None, *args, **kwargs):
        message = (message or self.__class__.__doc__) + self.EXTRA
        super(DjangoVersionMightBeUnsupported, self).__init__(message, *args, **kwargs)


class DjangoAdminNotInstalled(PyScaffoldDjangoError):
    """This extension depends on the ``django-admin`` cli script."""

    DEFAULT_MESSAGE = "django-admin is not installed, " "run: pip install django"

    def __init__(self, message=DEFAULT_MESSAGE, *args, **kwargs):
        super(DjangoAdminNotInstalled, self).__init__(message, *args, **kwargs)



import os
import shlex
import stat
import sys
import traceback
from os import environ
from pathlib import Path
from shutil import rmtree
from subprocess import STDOUT, CalledProcessError, check_output
from time import sleep
from uuid import uuid4
from warnings import warn

from pyscaffold.shell import get_executable

IS_POSIX = os.name == "posix"

PYTHON = sys.executable
"""Same python executable executing the tests... Hopefully the one inside the virtualenv
inside tox folder. If we install packages by mistake is not a huge problem.
"""


def uniqstr():
    """Generates a unique random long string every time it is called"""
    return str(uuid4())


def rmpath(path):
    """Carelessly/recursively remove path.
    If an error occurs it will just be ignored, so not suitable for every usage.
    The best is to use this function for paths inside pytest tmp directories, and with
    some hope pytest will also do some cleanup itself.
    """
    try:
        rmtree(str(path), onerror=set_writable)
    except FileNotFoundError:
        return
    except Exception:
        msg = f"rmpath: Impossible to remove {path}, probably an OS issue...\n\n"
        warn(msg + traceback.format_exc())


def set_writable(func, path, _exc_info):
    sleep(1)  # Sometimes just giving time to the SO, works

    if not Path(path).exists():
        return  # we just want to remove files anyway

    if not os.access(path, os.W_OK):
        os.chmod(path, stat.S_IWUSR)

    # now it either works or re-raise the exception
    func(path)


def merge_env(other=None, **kwargs):
    """Create a dict from merging items to the current ``os.environ``"""
    env = {k: v for k, v in environ.items()}  # Clone the environ as a dict
    env.update(other or {})
    env.update(kwargs)
    return env


def run(*args, **kwargs):
    """Run the external command. See ``subprocess.check_output``."""
    # normalize args
    if len(args) == 1:
        if isinstance(args[0], str):
            args = shlex.split(args[0], posix=IS_POSIX)
        else:
            args = args[0]

    if args[0] in ("python", "putup", "pip", "tox", "pytest", "pre-commit"):
        raise SystemError("Please specify an executable with explicit path")

    opts = dict(stderr=STDOUT, universal_newlines=True)
    opts.update(kwargs)

    try:
        return check_output(args, **opts)
    except CalledProcessError as ex:
        print("\n\n" + "!" * 80 + "\nError while running command:")
        print(args)
        print(opts)
        traceback.print_exc()
        msg = "\n******************** Terminal ($? = {}) ********************\n{}"
        print(msg.format(ex.returncode, ex.output))
        raise


def run_common_tasks(tests=True, docs=True, pre_commit=True, install=True):
    # Requires tox, setuptools_scm and pre-commit in setup.cfg ::
    # opts.extras_require.testing
    if tests:
        run(f"{PYTHON} -m tox")

    run(f"{PYTHON} -m tox -e build")
    wheels = list(Path("dist").glob("*.whl"))
    assert wheels

    run(f"{PYTHON} setup.py --version")

    if docs:
        run(f"{PYTHON} -m tox -e docs,doctests")

    if pre_commit:
        try:
            run(f"{PYTHON} -m pre_commit run --all-files")
        except CalledProcessError:
            print(run(get_executable("git"), "diff"))
            raise

    if install:
        assert Path(".venv").exists(), "Please use --venv when generating the project"
        venv_pip = get_executable("pip", prefix=".venv", include_path=False)
        assert venv_pip, "Pip not found, make sure you have used the --venv option"
        run(venv_pip, "install", wheels[0])



from pathlib import Path

import pytest
from pyscaffold import cli
from pyscaffold.file_system import chdir

from pyscaffoldext.django.extension import Django

from .helpers import run_common_tasks

EXT_FLAG = Django().flag


def test_add_custom_extension(tmpfolder):
    args = ["--no-config", EXT_FLAG, "my_project", "-p", "my_package"]
    # --no-config: avoid extra config from dev's machine interference
    cli.main(args)

    assert Path("my_project/src/my_package/__init__.py").exists()


@pytest.mark.slow
@pytest.mark.system
def test_generated_extension(tmpfolder):
    args = ["--no-config", "--venv", "--pre-commit", EXT_FLAG, "my_project"]
    # --no-config: avoid extra config from dev's machine interference
    # --venv: generate a venv so we can install the resulting project
    # --pre-commit: ensure generated files respect repository conventions
    cli.main(args)

    with chdir("my_project"):
        # Testing a project generated by the custom extension
        run_common_tasks()
